/*
  fixedpoint (c) 2006 Kris Beevers

  This file is part of fixedpoint.

  fixedpoint is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  fixedpoint is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with fixedpoint; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA

  $Id: fixedpoint.h,v 1.4 2007/02/06 03:27:05 beevek Exp $
*/

#ifndef _FIXEDPOINT_H
#define _FIXEDPOINT_H

#include <inttypes.h>
#ifdef __AVR__
#include <avr/pgmspace.h>
#else
#define PROGMEM
#define pgm_read_word(a) (*(a))
#define pgm_read_dword(a) (*(a))
#endif

#ifndef _USE_DOUBLE

#ifdef _USE_SHORT_FIXED_POINT

// NOT the default

// 6.10 format

// with this we get numbers up to 31.999023 and down to -32.000 with a
// resolution of 0.000977

typedef int16_t fixed_t;      // main fixed-point representation
typedef int32_t long_fixed_t; // used internally for mul/div
#define FP_DEC_BITS 10        // # bits used to represent the decimal part
#define FP_DEC_MASK 0x03ff    // to get the decimal part

#define fp_max_value   32767
#define fp_min_value   -32768

#define fp_1           1024   // 1
#define fp_1_2         512    // 1/2
#define fp_2           2048   // 2
#define fp_pi          3216   // pi
#define fp_pi_2        1608   // pi/2
#define fp_2_pi        6433   // 2pi
#define fp_e           2783   // e
#define fp_epsilon     2      // 0.002

// angular resolution of about 0.0078 radians or 0.4478 degrees;
// requires the storage of 402 values in the table for each function
#define fp_trig_shift   3     // every 8 angle values we have a table value

#define pgm_read_fixed pgm_read_word

#else // _USE_SHORT_FIXED_POINT

// the default

// 16.16 format

// with this we get a max of 32767.999985 with resolution 1.52588e-5.
// this will affect efficiency somewhat on the atmega64 (a 16-bit
// processor), though.

typedef int32_t fixed_t;      // main fixed-point representation
typedef int64_t long_fixed_t; // used internally for mul/div
#define FP_DEC_BITS 16        // # bits used to represent the decimal part
#define FP_DEC_MASK 0x0000ffff// to get the decimal part

#define fp_max_value   2147483647
#define fp_min_value   -2147483647

#define fp_1           65536    // 1
#define fp_1_2         32768    // 1/2
#define fp_2           131072   // 2
#define fp_pi          205887   // pi
#define fp_pi_2        102943   // pi/2
#define fp_2_pi        411774   // 2pi
#define fp_e           178145   // e
#define fp_epsilon     65       // 0.001

// almost the same angular resolution/storage requirements as above
#define fp_trig_shift  9        // every 512 angle values we have a table value

#define pgm_read_fixed pgm_read_dword

#endif // _USE_SHORT_FIXED_POINT

#else // _USE_DOUBLE

// although it's ultra slow, it may be useful to replace our
// computations with doubles or floats for accuracy comparision or
// overflow testing

#include <math.h>
#include <float.h>

typedef double fixed_t;
//typedef float fixed_t;

#define fp_max_value   DBL_MAX
#define fp_min_value   -DBL_MAX

#define fp_1           1.0    // 1
#define fp_1_2         0.5    // 1/2
#define fp_2           2.0    // 2
#define fp_pi          M_PI   // pi
#define fp_pi_2        M_PI_2 // pi/2
#define fp_2_pi        (2.0*M_PI) // 2pi
#define fp_e           M_E    // e
#define fp_epsilon     1e-5   // 0.00001

#define fp_trig_shift  9      // not actually used with doubles, here for compilation
#define pgm_read_fixed pgm_read_dword // also not used

#endif // _USE_DOUBLE

// function prototypes for fixed-point math

// conversions
fixed_t int2fp(int32_t i);
int32_t fp2int(fixed_t f);
fixed_t float2fp(float f);
float fp2float(fixed_t f);

// arithmetic
fixed_t fp_mul(fixed_t a, fixed_t b);
fixed_t fp_div(fixed_t a, fixed_t b);

// random numbers
void fp_srand(int seed);
fixed_t fp_rand(void);     // fp_min_value .. fp_max_value
fixed_t fp_rand_0_1(void); // 0 .. 1

// angles
fixed_t fp_clamp_angle(fixed_t a); // clamp a in -pi .. pi

// some standard math functionality
fixed_t fp_abs(fixed_t x);
fixed_t fp_min(fixed_t a, fixed_t b);
fixed_t fp_max(fixed_t a, fixed_t b);
fixed_t fp_floor(fixed_t x);
fixed_t fp_cos(fixed_t x);
fixed_t fp_sin(fixed_t x);
fixed_t fp_tan(fixed_t x);
fixed_t fp_atan(fixed_t x);
fixed_t fp_atan2(fixed_t y, fixed_t x);
fixed_t fp_sqrt(fixed_t x);
fixed_t fp_log(fixed_t x);
fixed_t fp_exp(fixed_t x);

// simple testing to determine the numerical error in a fixed-point
// operation
#define fp_error(op1,op2) (fp2float(op1)-(op2))

#endif // _FIXEDPOINT_H
