/*
  fixedpoint (c) 2006 Kris Beevers

  This file is part of fixedpoint.

  fixedpoint is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  fixedpoint is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with fixedpoint; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA

  $Id: fptest.c,v 1.1.1.1 2006/10/18 22:30:17 beevek Exp $
*/

#include <stdio.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "fixedpoint.h"

// generate the nth element of a Halton sequence with the given base
fixed_t halton(uint32_t n, uint32_t base)
{
  const fixed_t base_recip = fp_div(int2fp(1),int2fp(base));
  fixed_t n0 = int2fp(n), h = 0, f = base_recip, n1, r;
  while(n0 > 0) {
    n1 = fp_floor(fp_mul(n0,base_recip));
    r = n0 - fp_mul(n1, base);
    h += fp_mul(f, r);
    f = fp_mul(f, base_recip);
    n0 = n1;
  }
  return h;
}

float halton_f(uint32_t n, uint32_t base)
{
  const float base_recip = 1.0/(float)base;
  float n0 = n, h = 0, f = base_recip, n1, r;
  while(n0 > 0) {
    n1 = floor(n0*base_recip);
    r = n0 - n1*base;
    h += f*r;
    f = f*base_recip;
    n0 = n1;
  }
  return h;
}


int main()
{
  fixed_t a = int2fp(0), b = int2fp(-2), c = int2fp(1);
  printf("%d %d %d\n", a, b, c);
  printf("%f %f %f\n", fp2float(a), fp2float(b), fp2float(c));

  printf("b+c=%d\n", fp2int(b+c));
  printf("b*c=%d\n", fp2int(fp_mul(b,c)));
  printf("b/c=%f\n", fp2float(fp_div(b,c)));

  printf("%f -> %f\n", 3.141592, fp2float(float2fp(3.141592)));

  printf("sqrt(2) = %f\n", fp2float(fp_sqrt(int2fp(2))));
  printf("err(sqrt(2)) = %f\n", fp_error(fp_sqrt(int2fp(2)), sqrt(2.0)));

  printf("sqrt(0.01) = %f\n", fp2float(fp_sqrt(float2fp(0.01))));
  printf("err(sqrt(0.01)) = %f\n", fp_error(fp_sqrt(float2fp(0.01)), sqrt(0.01)));

  printf("max: %f\n", fp2float(fp_max_value));
  printf("min: %f\n", fp2float(fp_min_value));

  printf("halton float: ");
  uint32_t i;
  for(i = 0; i < 10; ++i)
    printf("%f ", (float)(halton_f(i, 5)));
  printf("\n");

  // note that this really doesn't work at all because of the
  // fixed-point imprecision!
  printf("halton fixed: ");
  for(i = 0; i < 10; ++i)
    printf("%f ", fp2float(halton(i, 5)));
  printf("\n");

  printf("random (-max..max): ");
  fp_srand(time(0));
  for(i = 0; i < 10; ++i)
    printf("%f ", fp2float(fp_rand()));
  printf("\n");
  printf("random (0..1): ");
  fp_srand(time(0));
  for(i = 0; i < 10; ++i)
    printf("%f ", fp2float(fp_rand_0_1()));
  printf("\n");

  printf("\nangle\t\tsin(a)\t\tcos(a)\t\ttan(a)\n");
  float ang;
  for(ang = -M_PI; ang < M_PI; ang += M_PI/16) {
    fixed_t f = float2fp(ang);
    printf("%f\t%f\t%f\t%f\n", ang, fp2float(fp_sin(f)), fp2float(fp_cos(f)), fp2float(fp_tan(f)));
  }

  printf("\nx\t\texp(x)\n");
  float x;
  for(x = -4.0; x < 3.0; x += 0.3) {
    fixed_t f = float2fp(x);
    printf("%f\t%f\n", x, fp2float(fp_exp(f)));
  }

  printf("\nx\t\tlog(x)\n");
  for(x = 0.0; x < 10.0; x += 0.1) {
    fixed_t f = float2fp(x);
    printf("%f\t%f\n", x, fp2float(fp_log(f)));
  }

  printf("\nx\t\tatan(x)\n");
  for(x = -7.0; x < 7.0; x += 0.75) {
    fixed_t f = float2fp(x);
    printf("%f\t%f\n", x, fp2float(fp_atan(f)));
  }

  printf("\nx\t\ty\t\tatan2(y,x)\tfloat-atan2(y,x)\terror\n");
  for(i = 0; i < 20; ++i) {
    fixed_t fx = fp_rand(), fy = fp_rand();
    fixed_t fpa = fp_atan2(fy, fx);
    float fa = atan2(fp2float(fy), fp2float(fx));
    printf("%f\t%f\t%f\t%f\t\t%f\n", fp2float(fx), fp2float(fy), fp2float(fpa), fa, fp_error(fpa,fa));
  }

  return 0;
}
